<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Cache
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

class Webtatic_Cache_Client
{
    
    /**
     * The last modified time of the page, as a unix timestamp.
     *
     * @var integer
     */
    protected $_lastModified = null;
    
    /**
     * The etag assigned to the page.
     *
     * @var string
     */
    protected $_eTag = null;
    
    /**
     * The time that the document will expire, as a unix timestamp.
     *
     * @var integer
     */
    protected $_expires = null;
    
    /**
     * The HTTP cache-control header to send.
     *
     * @var array
     */
    protected $_cacheControl = null;
    
    /**
     * Sets the page last modified date to the maximum value specified. If this
     * is set, then the "run" method will pass through successfully if it is
     * the later than the date the browser reports of the page, otherwise if
     * the other factors hold, it will tell the browser to use its cache.
     *
     * @param integer|array $time
     * @return Webtatic_Cache
     */
    public function lastModified($time)
    {
        if (!is_array($time)) {
            $time = func_get_args();
        }
        if ($this->_lastModified !== null) {
            $time[] = $this->_lastModified;
        }
        $this->_lastModified = max($time);
        if ($this->_expires === null) {
            $this->_expires = time();
        }
        
        return $this;
    }
    
    /**
     * Sets the page eTag. If this is set, then the "run" method will pass
     * through successfully if it does not match the browsers reported eTag of
     * the page, making the script recreate the cache, otherwise
     * if the other factors hold, it will tell the browser to use its cache.
     *
     * @param string $eTag
     * @return Webtatic_Cache
     */
    public function eTag($eTag)
    {
        $this->_eTag = $eTag;
        if ($this->_expires === null) {
            $this->_expires = time();
        }
        return $this;
    }
    
    /**
     * Sets the page's expiry date. If this is set, then the browser wont
     * ask the server for the document until the expiry date has passed.
     *
     * @param integer $time
     * @return Webtatic_Cache
     */
    public function expires($time)
    {
        $this->_expires = $time;
        return $this;
    }
    
    /**
     * Sets the cache control header. This is an array representation of the
     * values in the header.
     *
     * @param array $cacheControl
     * @return Webtatic_Cache
     */
    public function cacheControl($key, $value = true, $reset = false)
    {
        if (is_array($key)) {
            if ($reset) {
                $this->_cacheControl = null;
            }
            foreach ($key as $param => $value) {
                $this->cacheControl($param, $value, false);
            }
        } else {
            if ($this->_cacheControl === null) {
                $this->_cacheControl = array();
            } else if ($reset) {
                unset($this->_cacheControl[$key]);
            }
            if (is_bool($value)) {
                $this->_cacheControl[$key] = $value; 
            } else {
                $this->_cacheControl[$key][] = $value;
            }
        }
        return $this;
    }
    
    public function privateCache($header = null)
    {
        if (!is_array($header)) {
            $header = func_get_args();
        }
        if (count($header) == 0) {
            $this->cacheControl('private', true);
        } else {
            foreach ($header as $h) {
                $this->cacheControl('private', $h);
            }
        }
        return $this;
    }
    
    public function noCache($header = null)
    {
        if (!is_array($header)) {
            $header = func_get_args();
        }
        if (count($header) == 0) {
            $this->cacheControl('no-cache', true);
        } else {
            foreach ($header as $h) {
                $this->cacheControl('no-cache', $h);
            }
        }
        return $this;
    }
    
    protected function _formatCacheControl()
    {
        $array = array();
        foreach ($this->_cacheControl as $key => $value) {
            switch (gettype($value)) {
                case 'boolean':
                    if ($value) {
                        $array[] = rawurlencode($key);
                    }
                    break;
                case 'array':
                    foreach ($value as $v) {
                        if (is_integer($v)) {
                            $array[] = rawurlencode($key) . '=' . $v;
                        } else {
                            $array[] = rawurlencode($key) . '="' . $v . '"';
                        }
                    }
                    break;
            }
        }
        return join(',', $array);
    }
    
    /**
     * Checks the request headers for the browser cache variables, and tells
     * the browser to use its cache if they match, otherwise it carries on
     * letting PHP run the rest of the script. 
     *
     */
    public function run()
    {
        $ok_modified = $this->_lastModified === null || isset($_SERVER['HTTP_IF_MODIFIED_SINCE']);
        $ok_different_etag = $this->_eTag === null || isset($_SERVER['HTTP_IF_NONE_MATCH']);
        
        if ($this->_lastModified !== null || $this->_eTag !== null) {
            if ($ok_modified && $ok_different_etag) {
                $modified = isset($_SERVER['HTTP_IF_MODIFIED_SINCE']) && strtotime($_SERVER['HTTP_IF_MODIFIED_SINCE']) < $this->_lastModified;
                $different_etag = isset($_SERVER['HTTP_IF_NONE_MATCH']) && $_SERVER['HTTP_IF_NONE_MATCH'] != $this->_eTag;
                if (!$modified && !$different_etag) {
                    header('HTTP/1.1 304 Not Modified');
                    exit;
                }
            }
        }
        
        if ($this->_lastModified !== null) {
            header('Last-Modified: '.gmdate('D, d M Y H:i:s', $this->_lastModified) . ' GMT');
        }
        
        if ($this->_expires !== null) {
            header('Expires: '.gmdate('D, d M Y H:i:s', $this->_expires) . ' GMT');
        }
        
        if ($this->_eTag !== null) {
            header('ETag: "'.$this->_eTag.'"');
        }
        
        if ($this->_cacheControl !== null) {
            header('Cache-Control: '.$this->_formatCacheControl());
        }
    }
    
    /**
     * Creates a new instance of the Webtatic_Cache
     *
     * @return Webtatic_Cache
     */
    public static function factory()
    {
        return new Webtatic_Cache_Client();
    }
}