<?
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Db
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Db/Adapter/Interface.php';

class Webtatic_Db_Adapter_ZF implements Webtatic_Db_Adapter_Interface
{
    protected $_adapter;
    
    public function getAdapter() {
        return $this->_adapter;
    }
    
    public function __construct(Zend_Db_Adapter_Abstract $adapter)
    {
        $this->_adapter = $adapter;
    }
    
    public function __call($method, $args) {
        if (preg_match('/^select(.+)/i', $method, $match)) {
            $table = $this->getTable($match[1]);
            return $table->select();
        } else if (preg_match('/^find(.+)/i', $method, $match)) {
            $table = $this->getTable($match[1]);
            return call_user_func_array(array($table, 'find'), $args);
        } else if (preg_match('/^fetchNew(.+)/i', $method, $match)) {
            $table = $this->getTable($match[1]);
            return $table->fetchNew();
        }
        return call_user_func_array(array($this->_adapter, $method), $args);
    }
    
    /**
     * Updates table rows with specified data based on a WHERE clause.
     *
     * @param  mixed        $table  The table to update.
     * @param  array        $bind   Column-value pairs.
     * @param  mixed        $where  UPDATE WHERE clause(s).
     * @param  mixed        $params WHERE clause bind parameters.
     * @return int          The number of affected rows.
     */
    public function update($table, array $bind, $where = '', $params = null)
    {
        /**
         * Build "col = ?" pairs for the statement,
         * except for Zend_Db_Expr which is treated literally.
         */
        $set = array();
        foreach ($bind as $col => $val) {
            if ($val instanceof Zend_Db_Expr) {
                $val = $val->__toString();
                unset($bind[$col]);
            } else {
                $val = '?';
            }
            $set[] = $this->quoteIdentifier($col, true) . ' = ' . $val;
        }

        $where = $this->_whereExpr($where);

        /**
         * Build the UPDATE statement
         */
        $sql = "UPDATE "
             . $this->quoteIdentifier($table, true)
             . ' SET ' . implode(', ', $set)
             . (($where) ? " WHERE $where" : '');

        /**
         * Execute the statement and return the number of affected rows
         */
        $params = array_merge(array_values($bind), (array) $params);
        $stmt = $this->query($sql, $params);
        $result = $stmt->rowCount();
        return $result;
    }
    
    /**
     * Deletes table rows based on a WHERE clause.
     *
     * @param  mixed        $table  The table to update.
     * @param  mixed        $where  DELETE WHERE clause(s).
     * @param  mixed        $params WHERE clause bind parameters.
     * @return int          The number of affected rows.
     */
    public function delete($table, $where = '', $params = null)
    {
        $where = $this->_whereExpr($where);

        /**
         * Build the DELETE statement
         */
        $sql = "DELETE FROM "
             . $this->quoteIdentifier($table, true)
             . (($where) ? " WHERE $where" : '');

        /**
         * Execute the statement and return the number of affected rows
         */
        $stmt = $this->query($sql, (array) $params);
        $result = $stmt->rowCount();
        return $result;
    }
    
    /**
     * Convert an array, string, or Zend_Db_Expr object
     * into a string to put in a WHERE clause.
     *
     * @param mixed $where
     * @return string
     */
    protected function _whereExpr($where)
    {
        if (empty($where)) {
            return $where;
        }
        if (!is_array($where)) {
            $where = array($where);
        }
        foreach ($where as $cond => &$term) {
            // is $cond an int? (i.e. Not a condition)
            if (is_int($cond)) {
                // $term is the full condition
                if ($term instanceof Zend_Db_Expr) {
                    $term = $term->__toString();
                }
            } else {
                // $cond is the condition with placeholder,
                // and $term is quoted into the condition
                $term = $this->quoteInto($cond, $term);
            }
            $term = '(' . $term . ')';
        }

        $where = implode(' AND ', $where);
        return $where;
    }
    
    public function cloneAdapter()
    {
        $adapter = clone $this;
        $adapter->_connection = null;
        return $adapter;
    }
    
    public function getTable($table)
    {
        $table = ucfirst($table);
        Zend_Loader::loadClass($table);
        $table = new $table(array('db' => $this->_adapter));
        return $table;
    }
    
    public function getConnection() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getConfig() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function setProfiler($profiler) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getProfiler() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getStatementClass() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function setStatementClass($class) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function query($sql, $bind = array()) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function beginTransaction() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function commit() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function rollBack() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function insert($table, array $bind) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function select() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getFetchMode() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchAll($sql, $bind = array(), $fetchMode = null) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchRow($sql, $bind = array(), $fetchMode = null) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchAssoc($sql, $bind = array()) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchCol($sql, $bind = array()) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchPairs($sql, $bind = array()) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function fetchOne($sql, $bind = array()) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function quote($value, $type = null) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function quoteInto($text, $value, $type = null, $count = null) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function quoteIdentifier($ident, $auto=false) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function quoteColumnAs($ident, $alias, $auto=false) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function quoteTableAs($ident, $alias = null, $auto = false) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getQuoteIdentifierSymbol() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function lastSequenceId($sequenceName) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function nextSequenceId($sequenceName) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function foldCase($key) { return $this->__call(__FUNCTION__, func_get_args()); }
    
    public function listTables() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function describeTable($tableName, $schemaName = null) { return $this->__call(__FUNCTION__, func_get_args()); }
//    protected function _connect() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function isConnected() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function closeConnection() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function prepare($sql) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function lastInsertId($tableName = null, $primaryKey = null) { return $this->__call(__FUNCTION__, func_get_args()); }
//    protected function _beginTransaction() { return $this->__call(__FUNCTION__, func_get_args()); }
//    protected function _commit() { return $this->__call(__FUNCTION__, func_get_args()); }
//    protected function _rollBack() { return $this->__call(__FUNCTION__, func_get_args()); }
    public function setFetchMode($mode) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function limit($sql, $count, $offset = 0) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function supportsParameters($type) { return $this->__call(__FUNCTION__, func_get_args()); }
    public function getServerVersion() { return $this->__call(__FUNCTION__, func_get_args()); }
}