<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Db
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Zend/Db/Table/Abstract.php';
require_once 'Webtatic/Extender.php';

abstract class Webtatic_Db_Table_Abstract extends Zend_Db_Table_Abstract
{

    protected $_rowClass = 'Webtatic_Db_Table_Row';
    protected $_rowsetClass = 'Webtatic_Db_Table_Rowset';

    protected $_behaviours = array();

    protected $_behaviourObjects = null;

    protected $_rebuildStatements = array();
    protected $_rebuildStep = 0;
    protected $_rebuildTruncate = false;

    /* Returns an instance of a Zend_Db_Table_Select object.
     *
     * @return Webtatic_Db_Table_Select
     */
    public function select()
    {
        require_once 'Webtatic/Db/Table/Select.php';
        return new Webtatic_Db_Table_Select($this);
    }

    /**
     * @param string|array|Zend_Db_Table_Select $where  OPTIONAL An SQL WHERE clause or Zend_Db_Table_Select object.
     * @param string|array                      $order  OPTIONAL An SQL ORDER clause.
     * @return Zend_Db_Table_Row_Abstract The row results per the
     *     Zend_Db_Adapter fetch mode, or null if no row found.
     */
    public function fetchOne($where = null, $order = null, $bind = array())
    {
        if (!($where instanceof Zend_Db_Table_Select)) {
            $select = $this->select();

            if ($where !== null) {
                $this->_where($select, $where);
            }

            if ($order !== null) {
                $this->_order($select, $order);
            }

            $select->limit(1);

        } else {
            $select = $where->limit(1);
        }
        $rows = $this->_fetch($select, $bind);
        if (count($rows) == 0) {
            return null;
        }
        return array_shift($rows[0]);
    }

    public function count($where = null)
    {
        if (!($where instanceof Zend_Db_Table_Select)) {
            $select = $this->select();

            if ($where !== null) {
                $this->_where($select, $where);
            }

            if ($order !== null) {
                $this->_order($select, $order);
            }

            $select->limit(1);

        } else {
            $select = $where;
        }
        $select->assemble();
        $select->reset(Zend_Db_Select::COLUMNS)
        ->columns(array(new Zend_Db_Expr('COUNT(*)')));

        $rows = $this->_fetch($select);

        if (count($rows) == 0) {
            return null;
        }
        return array_shift($rows[0]);
    }

    public function loadTable($dependentTable)
    {
        if (is_string($dependentTable)) {
            try {
                @Zend_Loader::loadClass($dependentTable);
            } catch (Zend_Exception $e) {
                require_once 'Zend/Db/Table/Row/Exception.php';
                throw new Zend_Db_Table_Row_Exception($e->getMessage());
            }
            return new $dependentTable(array('db' => $this->getAdapter()));
        }
        return $table;
    }

    protected function _loadBehaviours()
    {
        $this->_behaviourObjects = array();
        $ext = Webtatic_Extender::get();
        if (!$ext->extensionPointExists('Webtatic_Db_Table_Behaviour')) {
            $ext->addExtensionPoint('Webtatic_Db_Table_Behaviour');
        }

        foreach ($this->_behaviours as $options) {
            $name = array_shift($options);
            $class = $ext->loadClass('Webtatic_Db_Table_Behaviour', $name);
            $behaviour = new $class($this, $options);
            $this->_behaviourObjects[] = $behaviour;
        }
    }

    public function callObservers($method, $args)
    {
        if ($this->_behaviourObjects === null) {
            $this->_loadBehaviours();
        }
        $called = false;
        foreach ($this->_behaviourObjects as $behaviour) {
            if (method_exists($behaviour, $method)) {
                call_user_func_array(array($behaviour, $method), $args);
                $called = true;
            }
        }

        //        if (!$called) {
        //            require_once 'Zend/Db/Table/Row/Exception.php';
        //            throw new Zend_Db_Table_Row_Exception("Unrecognized method '$method()'");
        //        }
    }

    /**
     * Fetches all rows.
     *
     * Honors the Zend_Db_Adapter fetch mode.
     *
     * @param string|array|Zend_Db_Table_Select $where  OPTIONAL An SQL WHERE clause or Zend_Db_Table_Select object.
     * @param string|array                      $order  OPTIONAL An SQL ORDER clause.
     * @param int                               $count  OPTIONAL An SQL LIMIT count.
     * @param int                               $offset OPTIONAL An SQL LIMIT offset.
     * @return Zend_Db_Table_Rowset_Abstract The row results per the Zend_Db_Adapter fetch mode.
     */
    public function fetchAll($where = null, $order = null, $count = null, $offset = null, $bind = array())
    {
        if (!($where instanceof Zend_Db_Table_Select)) {
            $select = $this->select();

            if ($where !== null) {
                $this->_where($select, $where);
            }

            if ($order !== null) {
                $this->_order($select, $order);
            }

            if ($count !== null || $offset !== null) {
                $select->limit($count, $offset);
            }

        } else {
            $select = $where;
        }

        $rows = $this->_fetch($select, $bind);

        $data  = array(
            'table'    => $this,
            'data'     => $rows,
            'readOnly' => $select->isReadOnly(),
            'rowClass' => $this->_rowClass,
            'stored'   => true,
            'foundRows' => $select->getCalcFoundRows() ?
                $this->getAdapter()->fetchOne('SELECT FOUND_ROWS()') : null
        );

        @Zend_Loader::loadClass($this->_rowsetClass);
        return new $this->_rowsetClass($data);
    }

    /**
     * Fetches one row in an object of type Zend_Db_Table_Row_Abstract,
     * or returns Boolean false if no row matches the specified criteria.
     *
     * @param string|array|Zend_Db_Table_Select $where  OPTIONAL An SQL WHERE clause or Zend_Db_Table_Select object.
     * @param string|array                      $order  OPTIONAL An SQL ORDER clause.
     * @return Zend_Db_Table_Row_Abstract The row results per the
     *     Zend_Db_Adapter fetch mode, or null if no row found.
     */
    public function fetchRow($where = null, $order = null, $bind = array())
    {
        if (!($where instanceof Zend_Db_Table_Select)) {
            $select = $this->select();

            if ($where !== null) {
                $this->_where($select, $where);
            }

            if ($order !== null) {
                $this->_order($select, $order);
            }

            $select->limit(1);

        } else {
            $select = $where->limit(1);
        }

        $rows = $this->_fetch($select, $bind);
        
        if (count($rows) == 0) {
            return null;
        }

        $data = array(
            'table'   => $this,
            'data'     => $rows[0],
            'readOnly' => $select->isReadOnly(),
            'stored'  => true
        );

        Zend_Loader::loadClass($this->_rowClass);
        return new $this->_rowClass($data);
    }

    protected function _fetch(Zend_Db_Table_Select $select, $bind = array())
    {
        $stmt = $this->_db->query($select, $bind);
        $data = $stmt->fetchAll(Zend_Db::FETCH_ASSOC);
        return $data;
    }

    protected function _rebuild($table)
    {

    }

    public function rebuild()
    {
        $db = $this->getAdapter();

        if ($this->_rebuildTruncate) {
            $db->query('TRUNCATE '.$db->quoteIdentifier($this->info(self::NAME)));
        }

        foreach ((array)$this->_rebuildStatements as $statement) {
            $offset = 0;
            do {
                $sql = $statement;
                if ($this->_rebuildStep > 0) {
                    $sql .= ' LIMIT '.$offset.', '.$this->_rebuildStep;
                }
                $sql = 'INSERT INTO '.$db->quoteIdentifier($this->info(self::NAME)). ' ' . $sql;
                $result = $db->query($sql);
                $offset += $result->rowCount();
            } while ($this->_rebuildStep == $result->rowCount());
        }

        $this->_rebuild($this->info(self::NAME));
    }
}
