<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Db
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Db/Table/Behaviour/Abstract.php';

class Webtatic_Db_Table_Behaviour_Nested
    extends Webtatic_Db_Table_Behaviour_Abstract
{

    const DIRECTION_UP = 'up';
    const DIRECTION_DOWN = 'down';
    const DIRECTION_LEFT = 'left';
    const DIRECTION_RIGHT = 'right';

    protected $_key = null;
    protected $_leftColumn = 'left';
    protected $_rightColumn = 'right';
    protected $_parentColumn = 'parent_id';

    public function getKey()
    {
        return $this->_key;
    }

    public function setKey($value)
    {
        $this->_key = (array) $value;

        return $this;
    }

    public function getLeftColumn()
    {
        return $this->_leftColumn;
    }

    public function setLeftColumn($value)
    {
        $this->_leftColumn = $value;

        return $this;
    }

    public function getRightColumn()
    {
        return $this->_rightColumn;
    }

    public function setRightColumn($value)
    {
        $this->_rightColumn = $value;

        return $this;
    }

    public function getParentColumn()
    {
        return $this->_parentColumn;
    }

    public function setParentColumn($value)
    {
        $this->_parentColumn = $value;

        return $this;
    }
    /**
     * Enter description here...
     *
     * @param Webtatic_Db_Table_Row_Abstract $row
     * @param bool $dirty
     */
    protected function _getKeyData($row, $useDirty = true)
    {
        return $row->getData($this->_key, $useDirty);
    }

    /**
     *
     * @param Webtatic_Db_Table_Row_Abstract $row
     * @param bool $dirty
     * @return Webtatic_Db_Table_Select $row
     */
    protected function _select($row, $useDirty = true)
    {
        $data = $this->_getKeyData($row, $useDirty);
        $adapter = $this->getTable()->getAdapter();
        $select = $this->getTable()->select();
        foreach ($data as $key => $value) {
            $select->where($adapter->quoteIdentifier($key) . ' = ?', $value);
        }

        return $select;
    }

    protected function _getParent($row, $useDirty = true)
    {
        $adapter = $this->getTable()->getAdapter();
        $parentId = $row->getData($this->_parentColumn, $useDirty);

        $primary = $this->getTable()->info('primary');
        $primaryColumn = $adapter->quoteIdentifier(array_shift($primary));

        return $this->_select($row, $useDirty)
            ->where("$primaryColumn = ?", $parentId)
            ->fetchRow();
    }

    public function moveAction($row, $direction)
    {
        $adapter = $this->getTable()->getAdapter();

        $parentColumn = $adapter->quoteIdentifier($this->_parentColumn);
        $leftColumn = $adapter->quoteIdentifier($this->_leftColumn);

        $parentId = $row->getData($this->_parentColumn, false);
        $left = $row->getData($this->_leftColumn, false);
        $right = $row->getData($this->_leftColumn, false);

        $select = $this->_select($row, false)
            ->where("$parentColumn = ?", $parentId);


        switch ($direction) {
            case 'up':
                $parent = $this->_getParent($row, false);
            case 'right':
                $sibling = $select
                    ->where("$leftColumn < ?", $left)
                    ->order("$this->_leftColumn DESC")
                    ->fetchRow();
                break;
            case 'down':
                $parent = $this->_getParent($row, false);
                $sibling = $select
                    ->where("$leftColumn > ?", $right)
                    ->order($this->_leftColumn)
                    ->limit(1, 1)
                    ->fetchAll();
                $sibling = $sibling->getRow(0);
                break;
            case 'left':
                $parent = $this->_getParent($row, false);

                $sibling = $this->_select($row, false)
                    ->where("$leftColumn > ?", $right)
                    ->where("$parentColumn = ?", $parent->getData($this->_parentColumn, false))
                    ->order($this->_leftColumn)
                    ->fetchRow();

                $parent = $this->_getParent($parent, false);
                break;
        }
        switch ($direction) {
            case 'up':
                if ($sibling) {
                    $this->insertBeforeAction($parent, $row, $sibling);
                }
                break;
            case 'down':
                $this->insertBeforeAction($parent, $row, $sibling);
                break;
            case 'right':
                if ($sibling) {
                    $this->appendAction($sibling, $row);
                }
                break;
            case 'left':
                if ($sibling) {
                    $this->insertBeforeAction($parent, $row, $sibling);
                } else {
                    $this->appendAction($parent, $row);
                }
        }
    }

    public function appendAction($parent, $row)
    {
        return $this->insertBeforeAction($parent, $row);
    }

    public function insertBeforeAction($parent, $row, $before = null)
    {

        if ($parent) {
            $primary = $this->getTable()->info('primary');
            $row->{$this->_parentColumn} = $parent->getData(array_shift($primary), false);
        } else {
            $row->{$this->_parentColumn} = 0;
        }
        if ($before) {
            $row->{$this->_leftColumn} = $before->getData($this->_leftColumn, false);
        } else {
            $row->{$this->_leftColumn} = null;
        }

//        var_dump($row->{$this->_parentColumn}, $row->{$this->_leftColumn}); exit;
        $row->save();

        return true;
    }

    public function insert($row)
    {
        $adapter = $this->getTable()->getAdapter();
//        $row->parent_id = 29;
        $rightColumn = $adapter->quoteIdentifier($this->_rightColumn);
        $parentColumn = $adapter->quoteIdentifier($this->_parentColumn);

        $data = $this->_getKeyData($row);
        $where = array();
        foreach ($data as $key => $value) {
            $where[] = $adapter->quoteIdentifier($key) . ' = ' . $adapter->quote($value);
        }

        if (!$row->getData($this->_parentColumn)) {
            $row->{$this->_parentColumn} = 0;
            $rightOf = $this->_select($row)
                ->from($this->getTable(), new Zend_Db_Expr("MAX($rightColumn)"))
                ->where("$parentColumn = 0")
                ->fetchOne();

            $row->{$this->_leftColumn} = $rightOf+1;
            $row->{$this->_rightColumn} = $rightOf+2;
        } else if ($parent = $this->_getParent($row)) {
            $this->_shiftNodes($where, $parent->right, 2);

            $row->{$this->_leftColumn} = $parent->right;
            $row->{$this->_rightColumn} = $parent->right+1;
        }
    }

    public function update($row)
    {
        $adapter = $this->getTable()->getAdapter();
        $data = $this->_getKeyData($row, false);
        $where = array();
        foreach ($data as $key => $value) {
            $where[] = $adapter->quoteIdentifier($key) . ' = ' . $adapter->quote($value);
        }

        $leftColumn = $adapter->quoteIdentifier($this->_leftColumn);
        $rightColumn = $adapter->quoteIdentifier($this->_rightColumn);
        $parentColumn = $adapter->quoteIdentifier($this->_parentColumn);

        $select = $this->_select($row, false)
            ->from($this->getTable(), new Zend_Db_Expr("MAX($rightColumn)"))
            ->where("$parentColumn = 0");


        $rightOf = $select
            ->fetchOne();

        $left = $row->getData($this->_leftColumn, false);
        $right = $row->getData($this->_rightColumn, false);

        if ($right == $rightOf) {
            $push = 0;
        } else {
            $push = 1 + $rightOf - $left;

            // shift selected nodes to the end of the index
            $this->_shiftNodes($where, $left, $push, $right+1);

            // TODO: updating the parent id to 0 here would maintain the index till they
            // are in the correct position, necissary?

            $push = $left - $right - 1;

            // shift all rows after the selected nodes back in place of the node
            $this->_shiftNodes($where, $left, $push);

            $p = $rightOf - $right;
            $left += $p;
            $right += $p;
        }
        $newLeft = $row->getData($this->_leftColumn);
        if ($newLeft == null) {
            if ($parent = $this->_getParent($row)) {
                $newLeft = $parent->getData($this->_rightColumn);
            } else {
                $newLeft = $left;

            }
        } else if ($newLeft > $row->getData($this->_leftColumn, false)) {
            $newLeft += $push;
        }

        if ($newLeft != $left) {
            $push = $right - $left + 1;

            // shift nodes after the new position to accomodate the selected nodes
            $this->_shiftNodes($where, $newLeft, $push);

            $left += $push;
            $right += $push;

            $push = $newLeft - $left;

            // shift the selected nodes into the new position
            $this->_shiftNodes($where, $left, $push);

            $left += $push;
            $right += $push;
        }

        $row->{$this->_leftColumn} = $left;
        $row->{$this->_rightColumn} = $right;
    }

    /**
     * Enter description here...
     *
     * @param Webtatic_Db_Table_Row_Abstract $row
     */
    public function delete($row)
    {
        $data = $this->_getKeyData($row, false);
        $adapter = $this->getTable()->getAdapter();

        $left = $row->getData($this->_leftColumn, false);
        $right = $row->getData($this->_rightColumn, false);

        $leftColumn = $adapter->quoteIdentifier($this->_leftColumn);
        $rightColumn = $adapter->quoteIdentifier($this->_rightColumn);

        $where = array();
        foreach ($data as $key => $value) {
            $where[] = $adapter->quoteIdentifier($key) . ' = ' . $adapter->quote($value);
        }

        $deleteWhere = array_merge($where, array(
            $adapter->quoteInto("$leftColumn > ?", $left),
            $adapter->quoteInto("$leftColumn < ?", $right),
        ));

        $adapter->delete($this->getTable()->info('name'), $deleteWhere);

        $this->_shiftNodes($where, $left, $left - $right - 1);
    }

    protected function _shiftNodes($where, $from, $count, $toExclusive = null)
    {
        $adapter = $this->getTable()->getAdapter();

        $leftColumn = $adapter->quoteIdentifier($this->_leftColumn);
        $rightColumn = $adapter->quoteIdentifier($this->_rightColumn);

        $updateWhere = array_merge($where, array(
            $adapter->quoteInto("$leftColumn >= ?", $from),
        ));

        if ($toExclusive) {
            $updateWhere[] = $adapter->quoteInto("$leftColumn < ?", $toExclusive);
        }

        $adapter->update($this->getTable()->info('name'), array(
            $this->_leftColumn => new Zend_Db_Expr($adapter->quoteInto("$leftColumn + ?", $count)),
        ), $updateWhere);

        $updateWhere = array_merge($where, array(
            $adapter->quoteInto("$rightColumn >= ?", $from),
        ));

        if ($toExclusive) {
            $updateWhere[] = $adapter->quoteInto("$rightColumn < ?", $toExclusive);
        }

        $adapter->update($this->getTable()->info('name'), array(
            $this->_rightColumn => new Zend_Db_Expr($adapter->quoteInto("$rightColumn + ?", $count)),
        ), $updateWhere);
    }
}