<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Extender
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

class Webtatic_Extender
{
    protected $_path = array();
    protected $_classNotLoaded = array();
    protected $_classLoaded = array();
    protected $_pathLoaded = array();
    protected $_extensionPoints = array();

    public function extensionPointExists($name)
    {
        return in_array($name, $this->_extensionPoints);
    }

    public function addExtensionPoint($name, $addDefault = true)
    {
        if (!$this->extensionPointExists($name)) {
            $this->_path[$name] = array();
            $this->_extensionPoints[] = $name;
            $this->_classLoaded[$name] = array();
            $this->_classNotLoaded[$name] = array();
            if ($addDefault) {
                $this->addExtensionPath($name, str_replace('_','/', $name));
            }
        }
    }

    public function addExtension($extensionPoint, $class, $params = null, $name = null, $path = null)
    {
        $this->addExtensionPoint($extensionPoint);
        if ($name === null) {
            if ($pos = strrpos($class, '_')) {
                $name = substr($class, $pos+1);
            } else {
                $name = $class;
            }
        }
        if ($path === null) {
            $path = str_replace('_','/', $class).'.php';
        }
        $this->_classNotLoaded[$extensionPoint][$name] = array('class' => $class, 'path' => $path, 'params' => $params);
    }

    public function addExtensionPath($extensionPoint, $path, $prefix = null)
    {
        $this->addExtensionPoint($extensionPoint);

        if ($prefix === null) {
            $prefix = $extensionPoint;
        }

        if (substr($prefix, -1) != '_') {
            $prefix .= '_';
        }
        foreach ((array) $path as $dir) {
            // attempt to strip any possible separator and
            // append the system directory separator
            $dir = str_replace(array('/', '\\'), DIRECTORY_SEPARATOR, $dir);
            $dir = rtrim($dir, DIRECTORY_SEPARATOR . DIRECTORY_SEPARATOR)
                 . DIRECTORY_SEPARATOR;

            // add as array with prefix and dir keys
            $this->_path[$extensionPoint][] = array('prefix' => $prefix, 'dir' => $dir);
        }
    }

    protected function _loadUnloadedExtensions($extensionPoint, $name = null)
    {
        if ($name !== null) {
            if (!array_key_exists($name, $this->_classNotLoaded[$extensionPoint])) {
                return;
            }
            $extensions = array(
                $name => $this->_classNotLoaded[$extensionPoint][$name]
            );
        } else {
            $extensions = $this->_classNotLoaded[$extensionPoint];
        }
        foreach ($extensions as $name => $extension) {
            $class = $extension['class'];
            if (class_exists($class, false)) {
                $this->_classLoaded[$extensionPoint][$name] = $extension;
            } elseif (self::isReadable($extension['path'])) {
                include_once $extension['path'];

                if (class_exists($class, false)) {
                    $this->_classLoaded[$extensionPoint][$name] = $extension;
                }
            }
            unset($this->_classNotLoaded[$extensionPoint][$name]);
        }
    }

    public function getExtensionPaths($extensionPoint)
    {
        return $this->_path[$extensionPoint];
    }

    public function getExtensions($extensionPoint)
    {
        $this->_loadUnloadedExtensions($extensionPoint);
        if (!$this->_pathLoaded[$extensionPoint]) {
            $path = get_include_path();
            $includeDirs = explode(PATH_SEPARATOR, $path);

            foreach (array_reverse($this->_path[$extensionPoint]) as $info) {
                $dir    = $info['dir'];
                $prefix = $info['prefix'];
                $glob = DIRECTORY_SEPARATOR.$dir.DIRECTORY_SEPARATOR.'*.php';

                foreach ($includeDirs as $includeDir) {
                    foreach (glob($includeDir.$glob) as $file) {
                        preg_match('#/([^/]*?)\.php$#', $file, $matches);
                        $name = ucfirst($matches[1]);
                        if (!array_key_exists($name, $this->_classLoaded[$extensionPoint])) {
                            $class = $prefix . $name;
                            if (class_exists($class, false)) {
                                $this->_classLoaded[$extensionPoint][$name] = $class;
                            } elseif (is_readable($file)) {
                                include_once $file;
                                if (class_exists($class, false)) {
                                    $this->_classLoaded[$extensionPoint][$name] = $class;
                                }
                            }
                        }
                    }
                }
            }
            $this->_pathLoaded[$extensionPoint] = true;
        }
        return $this->_classLoaded[$extensionPoint];
    }

    public function loadClass($extensionPoint, $name)
    {
        $name = ucfirst(trim($name));
        $this->_loadUnloadedExtensions($extensionPoint, $name);
        if (isset($this->_classLoaded[$extensionPoint][$name])) {
            return $this->_classLoaded[$extensionPoint][$name]['class'];
        }

        $file = $name . '.php';
        foreach (array_reverse($this->_path[$extensionPoint]) as $info) {
            $dir    = $info['dir'];
            $prefix = $info['prefix'];

            $class = $prefix . $name;
            if (class_exists($class, false)) {
                $this->_classLoaded[$extensionPoint][$name] = array(
                    'class' => $class,
                );
                return $class;
            } elseif (self::isReadable($dir . $file)) {
                include_once $dir . $file;
                if (class_exists($class, false)) {
                    $this->_classLoaded[$extensionPoint][$name] = array(
                        'class' => $class,
                    );
                    return $class;
                }
            }
        }
        throw new Exception("$type '$name' not found in path");
    }

    private static $instance = null;

    /**
     * @return Webtatic_Extender
     */
    public static function get()
    {
        if (self::$instance === null) {
            self::$instance = $instance = new Webtatic_Extender();
        } else {
            $instance = self::$instance;
        }
        return $instance;
    }

    public static function isReadable($filename)
    {
        if (@is_readable($filename)) {
            return true;
        }

        $path = get_include_path();
        $dirs = explode(PATH_SEPARATOR, $path);

        foreach ($dirs as $dir) {
            // No need to check against current dir -- already checked
            if ('.' == $dir) {
                continue;
            }

            if (@is_readable($dir . DIRECTORY_SEPARATOR . $filename)) {
                return true;
            }
        }

        return false;
    }
}
?>