<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Filesystem
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Parser/Replace.php';
require_once 'Webtatic/Filesystem/DirectoryIterator.php';
require_once 'Webtatic/Filesystem/Path.php';

class Webtatic_Filesystem_Glob
{
    protected $_path;
    protected $_params;
    protected $_defaults;
    protected $_patterns;
    
    protected $_iterator;
    protected $_data = null;
    protected $_matchIndex;
    
    public function getFiles()
    {
        return $this->_data;
    }
        
    
    public function __construct($patterns, $defaults = null, $reqs = null)
    {
        $this->compile($patterns, $defaults, $reqs);
    }
    
    protected function _quote($str)
    {
        return preg_quote($str, '#');
    }
    
    function glob_recursive($pattern, $path = null, $recurse = false)
    {
        $entries = array();
        if ($recurse) {
            foreach (glob($path.'/*', GLOB_BRACE | GLOB_ONLYDIR) as $entry ) {
                $entries_new = $this->glob_recursive($pattern, $entry, $recurse);
                array_splice($entries, count($entries), 0, $entries_new);
            }
        }
        $p = ($path === null ? '' : $path.'/').$pattern;
        if (preg_match('/^(?:(.*?)\/)?\*\*(?:(.*?))?$/', $p, $match)) {
            $glob_entries = $this->glob_recursive(isset($match[2]) ? '*'.$match[2]: '*', $match[1], true);
        } else {
            $glob_entries = glob($p, GLOB_BRACE);
        }
        return array_unique(array_merge($glob_entries, $entries));
    }
    
    public function compile($patterns, $defaults = null, $reqs = null)
    {
        if (is_string($patterns)) {
            $patterns = explode("\n", $patterns);
        }
        $this->_defaults = (array) $defaults;
        
        foreach ($patterns as $pattern) {
            if ($p = $this->_compilePattern($pattern, $reqs)) {
                $this->_patterns[] = $p;
            }
        }
    }
    
    protected function _compilePattern($pattern, $reqs = null)
    {
        $pattern = trim($pattern);
        if (strlen($pattern) > 0) {
            if ($pattern[0] == '+' || $pattern[0] == '-') {
                $exclude = $pattern[0] == '-';
                $pattern = trim(substr($pattern, 1));
            } else {
                $exclude = false;
            }
            $params = (object) array(
                'reqs' => (array)$reqs,
                'values' => array(),
            );
            $regexp = Webtatic_Parser_Replace::replace('#(.*?)(?:\{\:([^}]+)}|:([a-z]+)|{([^}]+)}|(\*+)|$)#', 
                array($this, '_compileReplace'), $pattern, $params);
            $p = preg_replace('#\{\:([^}]+)}|:([a-z]+)#', '*', $pattern);
            return array('regexp' => $regexp, 'p' => $p, 'params' => $params->values, 'exclude' => $exclude);
        }
        return null;
    }
    
    public function run($path = '.', $reqs = null)
    {
        $path = realpath($path);
        $files = array();
        $p = $this->_compilePattern($path, $reqs);
        foreach ($this->_patterns as $pattern) {
            $tfiles = $this->glob_recursive($pattern['p'], $p['p']);
            array_splice($files, count($files)-1, 0, $tfiles);    
        }
        $files = array_unique($files);
        $data = array();
        foreach ($files as $filename) {
            if (false !== ($result = $this->_matches($filename, $p['regexp'], $p['params']))) {
                $data[$result['matchIndex']][] = $result;
            }
        }
        if (count($data) > 0) {
            ksort($data); 
            return call_user_func_array('array_merge', $data);
        } else {
            return array();
        }
    }
    
    public function _compileReplace($match, $params)
    {
        $str = $this->_quote($match[1]);
        if (count($match) > 2 && ($match[2] || $match[3])) {
            $m = $match[2] ? $match[2] : $match[3];
            $params->values[] = $m;
            if (array_key_exists($m, $params->reqs)) {
                $str .= '('.$params->reqs[$m].')';
            } else {
                $str .= '([^'.preg_quote(DIRECTORY_SEPARATOR,'#').']+)';
            }
        } else if (count($match) > 4 && $match[4]) {
            $str .= '(?:'.str_replace(',', '|', $match[4]).')';
        } else if (count($match) > 5 && $match[5]){
            if (strlen($match[5]) == 1) {
                $str .= '[^'.preg_quote(DIRECTORY_SEPARATOR,'#').']*?';
            } else {
                $str .= '.*?';
            }
        }
        return $str;
    }
    
    public function matches($filename, $path = '.')
    {
        $path = Webtatic_Filesystem_Path::resolve($path);
        $pathexp = $this->_quote($path);
        return $this->_matches($filename, $pathexp);
    }
    
    public function _matches($filename, $pathexp, $pathParams = array())
    {
        $include = false;
        foreach ($this->_patterns as $key => &$pattern) {
            $regexp = $pattern['regexp'];
            if ($regexp[0] != DIRECTORY_SEPARATOR) {
                $regexp = $pathexp.$this->_quote('/').$regexp;
            }
            if (preg_match('#^'.$regexp.'$#', $filename, $match)) {
                if ($pattern['exclude']) {
                    $include = false;
                    unset($matchIndex);
                    unset($params);
                } else if (!isset($matchIndex)) {
                    $include = true;
                    $matchIndex = $key;
                    if (count($pathParams) > 0 || count($pattern['params']) > 0) {
                        array_shift($match);
                        $params = array_merge($this->_defaults, array_combine(array_merge($pathParams, $pattern['params']), $match));
                    } else {
                        $params = $this->_defaults;
                    }
                }
            }
        }
        $params['filename'] = $filename;
        $params['matchIndex'] = $matchIndex;
        return $include ? $params : false;
    }
    
    public static function match($filename, $patterns, $path = '.', $defaults = null, $reqs = null)
    {
        if (!$patterns instanceof Webtatic_Filesystem_Glob) {
            $patterns = new Webtatic_Filesystem_Glob($patterns, $path, $defaults, $reqs); 
        }
        return $patterns->matches($filename, $path);
    }
}
?>