<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Filter
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Parser/Abstract.php';

class Webtatic_Filter_BBCodeToHTML extends Webtatic_Parser_Abstract
{
    
    protected $_defaultState = 'standard';
    protected $_missCallback = 'escape';
    protected $_patterns = array(
        'inline' => array(
            // inline tags
            array('inlineStart' => '\[(b|i|u)\]'),
            array('tagEnd' => '\[/([^\]]*)\]'),
            array('colorStart' => '\[colou?r=([^\]]*)\]'),
            array('sizeStart' => '\[size=([^\]]*)\]'),
            array('email' => '\[email]([^\s\[]+)(?:\[/email\])?'),
            array('url' => '\[url](([a-z]+://)?[^\s\[]+)(?:\[/url\])?'),
            array('urlAndLabelStart' => '\[url=(([a-z]+://)?[^\]]*)]'),
            array('image' => '\[img\]([^\s\[]+)(?:\[/img\])?'),

            // magic links
            array('email' => '([^\s\[@]+@[^\s\[]+)'),
            array('url' => '(www\.[^\s\r\n]+)'),
            array('url' => '(([a-z]+://)[^\s]+)'),
        ),
        'standard' => array(
            array('#inline', '#standard'),
            
            // block tags
            array('quoteStart' => '\[quote(?:="([^"]*)")?\]\s*', '#standard'),
            array('listStart' => '\[list(?:=(a|1))?]', '#list'),
            array('code' => '\[code(?:=([^\]]+))?](.*?)(?:\[/code\]|$)', '#standard'),
    
            array('passthrough' => '[ \t]+', '#standard'),
            array('newLine' => '(\r\n?|\n)+', '#standard'),
            array('escape' => '[^\s\[]+', '#standard'),
        ),
        'list' => array(
            array('skip' => '\s+', '#listItem'),
            array('listItemStart' => '\[\*\]', '#listItem'),
            array('listEnd' => '\[/(list)\]'),
        ),
        'listItem' => array(
            array('#inline', '#listItem'),
            array('passthrough' => '[ \t]+', '#listItem'),
            array('listItemEnd' => '(\r\n?|\n+)'),
        ),
    );
    
    protected $_stack = array();
    
    protected $_precedence = array('quote', 'block', 'inline');
    
    public function __construct($config = array())
    {
        $config = (array) $config;
        if (!array_key_exists('encoding', $config)) {
            $config['encoding'] = 'UTF-8';
        }
        parent::__construct($config);
    }

    public function filter($content)
    {
        return $this->_parse($content).$this->pop();
    }
    
    protected function _escape($value)
    {
        return htmlentities($value, ENT_COMPAT, $this->_config['encoding']);
    }
    
    protected function push($callback, $tag)
    {
        if (!method_exists($this, $callback.'Start') || call_user_func(array($this, $callback.'Start'), $tag)) { 
            $this->_stack[] = array($callback, $tag);
            return true;
        }
        return false;
    }
    
    protected function pop($type = null, $tag = null)
    {
        $html = '';
        $p = array_search($type, $this->_precedence);
        while ($t = array_pop($this->_stack)) {
            $currentP = array_search($t[0], $this->_precedence);
            if ($currentP < $p) {
                array_push($this->_stack, $t);
                break;
            }
            $html .= call_user_func(array($this, $t[0].'End'), $t[1]);
            if ($t[1] == $tag) {
                break;
            }
        }
        return $html;
    }
    
    protected function inline($tag = null)
    {
        $html = '';
        if ($this->_state == 'standard' && $this->push('block', 'p')) {
            $html .= '<p>';
        }
        if ($tag !== null && $this->push('inline', $tag)) {
            $html .= '<'.$tag.'>';
        }
        return $html;
    }
    
    protected function formatEscape($match)
    {
        return $this->inline().$this->_escape($match[0][0]);
    }
    
    protected function formatInlineStart($match)
    {
        return $this->inline($match[1][0]);
    }
    
    protected function formatTagEnd($match)
    {
        if ($match[1][0] == 'quote') {
            $type = 'quote';
        } else {
            $type = 'inline';
        }
        return $this->pop($type, $match[1][0]);
    }
    
    protected function inlineStart($tag)
    {
        foreach ($this->_stack as $t) {
            if ($t[1] == $tag) {
                return false;
            }
        }
        return true;
    }
    
    protected function inlineEnd($tag)
    {
        if ($tag == 'size' || $tag == 'color' || $tag == 'colour') {
            return '</span>';
        }
        if ($tag == 'url') {
            return '</a>';
        }
        return '</'.$tag.'>';
    }
    
    protected function blockStart($tag)
    {
        return $this->inlineStart($tag);
    }
    
    protected function blockEnd($tag)
    {
        return '</'.$tag.'>';
    }
    
    protected function formatQuoteStart($match)
    {
        $html = $this->pop('block');
        $this->push('quote', 'quote');
        if ($match[1][0]) {
            return $html.'<div class="quote">'.$this->_escape($match[1][0]).' says:<div class="content">';
        } else {
            return $html.'<div class="quote">Quote:<div class="content">';
        }
    }
    
    protected function quoteEnd($tag)
    {
        return '</div></div>';
    }
    
    protected function formatColorStart($match)
    {
        $html = $this->inline();
        if ($this->push('inline', 'color')) {
            $html .= '<span style="color: '.$this->_escape($match[1][0]).'">';
        }
        return $html;
    }
    
    protected function formatSizeStart($match)
    {
        $html = $this->inline();
        if ($this->push('inline', 'size')) {
            $html .= '<span style="font-size: '.$this->_escape($match[1][0]).'%">';
        }
        return $html;
    }
    
    protected function formatUrlAndLabelStart($match)
    {
        $html = $this->inline();
        $this->push('inline', 'url');
        $url = !$match[2][0] ? 'http://'.$match[1][0] : $match[1][0];
        return $html.'<a href="'.$this->_escape($url).'">';
    }
    
    protected function formatUrl($match)
    {
        $html = $this->inline();
        $url = !$match[2][0] ? 'http://'.$match[1][0] : $match[1][0];
        return $html.'<a href="'.$this->_escape($url).'">'.$this->_escape($match[1][0]).'</a>';
    }
    
    protected function formatEmail($match)
    {
        return $this->inline().'<a href="mailto:'.$this->_escape($match[1][0]).'">'.$this->_escape($match[1][0]).'</a>';
    }
    
    protected function formatCode($match)
    {
        $html = $this->pop('block');
        
        if ($match[1][0]) {
            require_once 'Webtatic/Filter/CodeToHTML.php';
            $filter = new Webtatic_Filter_CodeToHTML(array('language' => $match[1][0]));
            $code = $filter->filter($match[2][0]);
        } else {
            $code = '<pre>'.$this->_escape($match[2][0]).'</pre>';
        }
        return $html.$code;
    }
    
    protected function formatImage($match)
    {
        return $this->inline().'<img src="'.$this->_escape($match[1][0]).'" alt="" />';
    }
    
    protected function formatNewLine($match)
    {
        $match[0][0] = str_replace("\r", '', $match[0][0]);
        if (($c = strlen($match[0][0])) > 1) {
            $html = str_repeat('<br />', $c-2) . $this->pop('block');
        } else {
            $html = '<br />';
        }
        return $html;
    }
    
    protected function formatListStart($match)
    {
        $html = $this->pop('block');
        if ($match[1][0]) {
            if ($match[1][0] == '1') {
                $style = 'decimal';
            } else {
                $style = 'lower-alpha';
            }
            $html .= '<ol style="list-style-type: '.$style.'">';
            $this->push('block', 'ol');
        } else {
            $html .= '<ul>';
            $this->push('block', 'ul');
        }
        
        return $html;
    }
    
    protected function formatListEnd($match)
    {
        return $this->pop('block');
    }
    
    protected function formatListItemStart($match)
    {
        return $this->inline('li');
    }
    
    protected function formatListItemEnd($match)
    {
        return $this->pop('inline', 'li');
    }
}