<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Filter
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Parser/Abstract.php';

class Webtatic_Filter_MediaWikiToHTML extends Webtatic_Parser_Abstract
{
    
    protected $_defaultState = 'standardSOL';
    protected $_missCallback = 'escape';
    
    protected $_patterns = array(
        'inline' => array(
            array('quotes' => "'{2,}"),
            array('externalLinkMagic' => '((?:([a-z]+://)|www\.)[^\s\r\n]+)'),
            array('externalLink' => '\[(([a-z]+://)?[^\s]+)\s+([^\]]+)\]'),
//            "\[\[([^\|\]*)(?:\|([^\]*)])?\]\]" => array(2, 'internalLink'),
        ),
        'standardSOL' => array(
            array('#inline', '#standard'),
            array('headingStart' => "={2,}", '#heading'),
            array('preLine' => " ([^\r\n]*)(?:\r\n?|\n)?", '#standardSOL'),
            array('hr' => "----", '#standardSOL'),
            array('newLine' => "\r\n?|\n", '#standardSOL'),
            array('listLineStart' => '([*\#]+) *', '#listItem'),
            array('otherSOL' => ".", '#standard'),
        ),
        'standard' => array(
            array('#inline', '#standard'),
            array('newLineStandard' => "\r\n?|\n", '#standardSOL'),
        ),
        'heading' => array(
            array('#inline', '#heading'),
            array('headingEnd' => "={2,}", '#heading'),
            array('newLineHeading' => "\r\n?|\n"),
        ),
        'listItem' => array(
            array('#inline', '#listItem'),
            array('listItemEnd' => "\r\n?|\n"),
        ),
    );
    
    protected $_stack = array();
    
    protected $_precedence = array('block', 'inline');
    
    public function filter($content)
    {
        return $this->_parse($content).$this->pop();
    }
    
    protected function toggle($type, $tag, $toggle = null)
    {
        foreach ($this->_stack as $item) {
            if ($item == array($type, $tag)) {
                if ($toggle === null || !$toggle) {
                    return $this->pop($type, $tag);
                } else {
                    return '';
                }
            }
        }
        
        if ($toggle === null || $toggle) {
            return $this->push($type, $tag);
        } else {
            return '';
        }
    }
    
    protected function _escape($value)
    {
        return htmlentities($value, ENT_COMPAT, $this->_config['encoding']);
    }
    
    protected function formatEscape($match)
    {
        return $this->inline().$this->_escape($match[0][0]);
    }
    
    public function formatOtherSOL($match)
    {
        return $this->formatEscape($match);
    }
    
    protected function inline($tag = null)
    {
        $html = '';
        if ($this->_state == 'standardSOL') {
            $html .= $this->toggle('block', 'p', true);
        }
        return $html;
    }
    
    protected function push($callback, $tag)
    {
        if ($html = call_user_func(array($this, $callback.'Start'), $tag)) {
            $this->_stack[] = array($callback, $tag);
            return $html;
        }
        return false;
    }
    
    protected function pop($type = 'block', $tag = null)
    {
        $html = '';
        $p = array_search($type, $this->_precedence);
        $push = array();
        while ($t = array_pop($this->_stack)) {
            $currentP = array_search($t[0], $this->_precedence);
            if ($currentP < $p) {
                array_push($this->_stack, $t);
                break;
            }
            $html .= call_user_func(array($this, $t[0].'End'), $t[1]);
            if ($t[1] == $tag) {
                foreach ($push as $t) {
                    $html .= $this->push($t[0], $t[1]);
                }
                break;
            }
            if ($currentP == $p) {
                array_unshift($push, $t);
            }
        }
        return $html;
    }
    
    public function formatQuotes($match)
    {
        $count = strlen($match[0][0]);
        $html = '';
        if ($count > 2) {
            $html .= $this->toggle('inline', 'b');
            $count -= 3;
        }
        if ($count > 1) {
            $html .= $this->toggle('inline', 'i');
            $count -= 2;
        }
        return str_repeat("'", $count) . $html;
    }
    
    public function formatNewLineStandard()
    {
        return ' ';
    }
    
    public function formatNewLine()
    {
        return $this->pop('block');
    }
    
    public function formatNewLineHeading()
    {
        return $this->pop('block');
    }
    
    public function formatParagraph()
    {
        $html = '';
        if ($this->_state == 'standard') {
            $html .= ' ';
        } else {
            $html .= $this->pop('block');
        }
        return $html;
    }
    
    public function inlineStart($tag)
    {
        return $this->inline().'<'.$tag.'>';
    }
    
    public function inlineEnd($tag)
    {
        return '</'.$tag.'>';
    }
    
    public function blockStart($tag)
    {
        return '<'.$tag.'>';
    }
    
    public function blockEnd($tag)
    {
        return '</'.$tag.'>';
    }
    
    public function formatHr($match)
    {
        return $this->pop() . '<hr />';
    }
    
    public function formatHeadingStart($match)
    {
        return $this->pop().$this->push('block', 'h'.strlen($match[0][0]));
    }
    
    public function formatHeadingEnd($match)
    {
        foreach ($this->_stack as $f) {
            if ($f[0] == 'block' && $f[1][0] == 'h') {
                if (substr($f[1], 1) == strlen($match[0][0])) {
                    return $this->pop('block', $f[1]);
                } else {
                    return $this->_escape($match[0][0]);
                }
            }
        }
        return $this->pop();
    }
    
    public function formatPreLine($match)
    {
        return $this->toggle('block', 'pre', true).$this->_escape($match[1][0])."\n";
    }
    
    public function formatExternalLinkMagic($match)
    {
        return $this->formatExternalLink($match);
    }
    
    public function formatExternalLink($match)
    {
        if (count($match) == 2 || $match[2][0] == '') {
            $url = 'http://'.$match[1][0];
        } else {
            $url = $match[1][0]; 
        }
        
        if (count($match) < 4) {
            $label = $url;
        } else {
            $label = $match[3][0];
        }
        return '<a href="'.$this->_escape($url).'">'.$this->_escape($label).'</a>';
    }
    
    public function stackCmp($a, $b)
    {
        return strcmp($a[0], $b[0]) || strcmp($a[1], $b[1]);
    }
    
    public function formatListLineStart($match)
    {
        $html = '';
        
        $stack = $this->_stack;
        $newStack = array();
        foreach (str_split($match[1][0]) as $letter) {
            $newStack[] = array('block', $letter == '#' ? 'ol' : 'ul');
            $newStack[] = array('block', 'li');
        }
        
        array_pop($newStack);
        
        $pop = array_udiff_assoc($stack, $newStack, array($this,'stackCmp'));
        $push = array_udiff_assoc($newStack, $stack, array($this,'stackCmp'));
        
        foreach (array_reverse($pop) as $t) {
            $html .= $this->pop($t[0], $t[1]);
        }
        
        foreach ($push as $t) {
            $html .= $this->push($t[0], $t[1]);
        }
        
        $html .= $this->push('block', 'li');
        return $html;
    }
    
    public function formatListItemEnd($match)
    {
        return '';
    }
}
