<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Form
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Form/Field.php';
require_once 'Webtatic/Form/Exception.php';
require_once 'Webtatic/Form/Field/Interface.php';
require_once 'Webtatic/Form/Rule.php';

class Webtatic_Form_FieldSet implements Webtatic_Form_Field_Interface
{
    protected $_parent;

    protected $_fields;
    protected $_rules = array();
    protected $_conditions = array();
    protected $_remember = true;

    protected $_multiple = false;

    protected $_name;
    protected $_label = null;
    protected $_attributes = array();

    protected $_fieldClass = 'scalar';
    protected $_ruleClass = 'Webtatic_Form_Rule';

    protected $_template = null;

    protected $_valid = null;
    protected $_errors = array();
    protected $_messages = array();

    protected $_items = null;
    protected $_index = null;

    public function __clone()
    {
        foreach ($this->_fields as $key => $value) {
            $this->_fields[$key] = clone $value;
            $this->_fields[$key]->setParent($this);
        }

        foreach ($this->_rules as $key => $value) {
            $this->_rule[$key] = clone $value;
        }
    }

    public function __construct($name, $options = array(), $parent = null)
    {
        $this->_name = $name;
        $this->_parent = $parent;
        $this->setOptions($options);
    }

    public function setOption($name, $value)
    {
        $method = 'set'.ucfirst($name);
        if (method_exists($this, $method)) {
            return $this->$method($value);
        } else {
            return $this->__set($name, $value);
        }
    }

    public function getLabel()
    {
        return $this->_label ? $this->_label : $this->getName();
    }

    public function setLabel($value)
    {
        $this->_label = $value;

        return $this;
    }

    public function getAttributes()
    {
        return $this->_attributes;
    }

    public function setAttributes($value)
    {
        $this->_attributes = $value;
        return $this;
    }

    public function getParent()
    {
        return $this->_parent;
    }

    public function setParent($value)
    {
        $this->_parent = $value;

        return $this;
    }

    public function getForm()
    {
        if ($this->_parent == null) {
            return $this;
        } else {
            return $this->getParent()->getForm();
        }
    }

    public function getRemember()
    {
        return $this->_remember;
    }

    public function setRemember($value) {
        $this->_remember = $value;
    }

    /**
     * Enter description here...
     *
     * @param array $options
     * @return Webtatic_Form
     */
    public function setOptions($options)
    {
        foreach ((array) $options as $key => $value) {
            $this->setOption($key, $value);
        }

        return $this;
    }

    public function getName()
    {
        return $this->_name;
    }

    public function getId()
    {
        $parentName = $this->getParent() ? $this->getParent()->getId() : null;
        if ($parentName != '') {
            $name = $parentName . '_' . $this->getName();
        } else {
            $name = $this->getName();
        }

        if ($this->_index !== null) {
            $name .= '_' . $this->_index;
        }
        return $name;
    }

    public function getFullName()
    {
        $parentName = $this->getParent() ? $this->getParent()->getFullName() : null;
        if ($parentName != '') {
            $name = $parentName . '[' . $this->getName() . ']';
        } else {
            $name = $this->getName();
        }

        if ($this->_index !== null) {
            $name .= '[' . $this->_index . ']';
        }

        return $name;
    }

    public function reset()
    {
        $this->loadData(null, true);
    }
    
    protected function _load()
    {
        
    }
    
    protected function _postLoad()
    {
        
    }

    public function loadData($data, $default = false)
    {
        $this->_preLoad();
        
        if ($data) {
            if (is_object($data) && method_exists($data, 'toArray')) {
                $data = $data->toArray();
            } 
            
            $d = array();
            foreach ($data as $key => $value) {
                list($key1, $key2) = explode('_', $key, 2);
                if ($key2) {
                    $d[$key1][$key2] = $value;
                } else {
                    $d[$key1] = $value;
                }
            }
            $data = $d;
        }
        
        if ($this->getMultiple()) {
            $this->_items = array();
            if ($default) {
                for ($i = 0; $i < $this->getMultiple(); $i++) {
                    $this->_items[] = $item = clone $this;
                    $item->setMultiple(false);
                    $item->_index = $i;
                    $item->loadData(null, true);
                }
            } else {
                foreach ($data as $i => $d) {
                    $this->_items[] = $item = clone $this;
                    $item->setParent($this);
                    $item->setMultiple(false);
                    $item->_index = $i;
                    $item->loadData($d, false);
                }
            }
        } else {
            foreach ((array) $this->getFields() as $key => $field) {
                if ($field instanceof Webtatic_Form_FieldSet) {
                    if (is_object($data)) {
                        $field->loadData($data->$key, $default);
                    } else {
                        $field->loadData($data[$key], $default);
                    }
                } else {
                    if (!$default && !$field->getReadOnly()) {
                        if (is_object($data)) {
                            $field->setRawValue($data->$key);
                        } else {
                            $field->setRawValue($data[$key]);
                        }
                    } else {
                        $field->reset();
                    }
                }
            }
        }
        
        $this->_postLoad();
    }

    public function getItems()
    {
        if ($this->getMultiple()) {
            return $this->_items;
        } else {
            return array($this);
        }
    }

    public function isValid($partial = false)
    {
        if ($this->_valid === null) {
            $this->process($partial);
        }

        return $this->_valid;
    }

    public function setValid($value)
    {
        $this->_valid = $value;

        return $this;
    }

    public function getErrors($partial = false)
    {
        if ($this->_valid === null) {
            $this->process($partial);
        }

        return $this->_errors;
    }

    public function addMessage($key, $message)
    {
        $this->_messages[$key] = $message;
    }

    public function setErrors($value)
    {
        $this->_errors = $value;

        return $this;
    }

    public function getMessages($partial = false)
    {
        if ($this->_valid === null) {
            $this->process($partial);
        }

        return $this->_messages;
    }

    public function setMessages($value)
    {
        $this->_messages = $value;

        return $this;
    }

    public function addMessages($value)
    {
        $this->_messages = array_merge($this->_messages, $value);

        return $this;
    }

    public function toArray($filtered = true)
    {
        if ($filtered && $this->_valid === null) {
            $this->process();
        }
        $d = array();
        if ($this->getMultiple()) {
            foreach ($this->_items as $i => $item) {
                $d[$i] = $item->toArray($filtered);
            }
        } else {
            foreach ((array) $this->_fields as $key => $field) {
                if (is_array($data)) {
                    $d[$key] = array();
                    foreach ($data as $c) {
                        $d[$key][] = $c->toArray($filtered);
                    }
                } else if ($field instanceof Webtatic_Form_FieldSet) {
                    $d[$key] = $field->toArray($filtered);
                } else if ($filtered) {
                    $d[$key] = $field->getValue();
                } else {
                    $d[$key] = $field->getRawValue();
                }
            }
        }
        return $d;
    }

    public function getValueOptions()
    {
        return null;
    }

    public function getMultiple()
    {
        return $this->_multiple;
    }

    public function setMultiple($value)
    {
        $this->_multiple = $value;

        return $this;
    }

    public function addCondition($condition)
    {
        if (is_object($condition)) {
            if ($condition instanceof Webtatic_Form_Rule) {
                $this->_conditions[] = $condition;
            } else {
                throw new Webtatic_Form_Exception('Object of class "'.classname($condition).'" is not extended from "Webtatic_Form_Rule"');
            }
        } else if (is_bool($condition)) {
            $this->_conditions[] = $condition;
        } else {
            $this->_conditions[] = new Webtatic_Form_Rule($this, 'conditions', $condition);
        }

        return $this;
    }

    public function setConditions($conditions)
    {
        foreach ($conditions as $condition) {
            $this->addCondition($condition);
        }

        return $this;
    }

    public function render($view, $template = null)
    {
        if (!$template) {
            $template = $this->getTemplate();
        }
        if ($this->getMultiple()) {
            $html = '';
            foreach ((array)$data as $key => $value) {
                $html .= $view->partial($template, array('form' => $this));
            }
            return $html;
        } else {
            return $view->partial($template, array('form' => $this));
        }
    }

    public function getTemplate()
    {
        return $this->_template !== null ? $this->_template : $this->getParent()->getTemplate();
    }

    public function setTemplate($value)
    {
        $this->_template = $value;

        return $this;
    }

    /**
     * Enter description here...
     *
     * @param string|Webtatic_Form_Field_Interface $name
     * @param array $options
     * @return Webtatic_Form
     */
    public function setField($name, $options = null)
    {
        if (is_object($name)) {
            if ($name instanceof Webtatic_Form_Field_Interface) {
                $this->_fields[$name->getName()] = $field = $name;
            } else {
                throw new Webtatic_Form_Exception('Object of class "'.get_class($name).'" does not implement "Webtatic_Form_Field_Interface"');
            }
        } else {
            list($name, $subName) = explode('_', $name, 2);

            if ($subName) {
                if (!isset($this->_fields[$name])) {
                    $this->_fields[$name] = new Webtatic_Form_Fieldset($name, null, $this);
                }
                return $this->_fields[$name]->setField($subName, $options);
            } else if (is_object($options)) {
                if ($options instanceof Webtatic_Form_Field_Interface) {
                    $this->_fields[$name] = $field = $options;
                    $options = array();
                } else {
                    throw new Webtatic_Form_Exception('Object of class "'.get_class($options).'" does not implement "Webtatic_Form_Field_Interface"');
                }
            } else {
                if (! ($field = $this->getField($name))) {
                    switch ($options['type']) {
                        case 'fieldSet':
                            require_once 'Webtatic/Form/FieldSet.php';
                            $class = 'Webtatic_Form_FieldSet';
                            break;
                        case '':
                            $options['type'] = $this->_fieldClass;
                        default:
                            try {
                                $class = Webtatic_Form::loadClass('Field', $options['type']);
                            } catch (Exception $e) {
                                $class = $this->_fieldClass;
                            }
                    }
                    unset($options['type']);
                    $this->_fields[$name] = $field = new $class($name, array());
                }
            }
        }
        $field->setOptions($options);
        $field->setParent($this);

        return $this;
    }

    public function getIndex()
    {
        return null;
    }

    public function __get($name)
    {
        list($name, $subName) = explode('_', $name, 2);
        if ($subName) {
            return $this->_fields[$name]->__get($subName);
        } else {
            return $this->getField($name);
        }
    }

    public function __isset($name)
    {
        return isset($this->_fields[$name]);
    }

    public function __set($name, $value)
    {
        list($name, $subName) = explode('_', $name, 2);
        if ($subName) {
            if (!isset($this->_fields[$name])) {
                $this->_fields[$name] = new Webtatic_Form_Fieldset($name,array(), $this);
            }
            return $this->_fields[$name]->__set($subName, $value);
        } else {
            return $this->setField($name, $value);
        }
    }

    /**
     * Enter description here...
     *
     * @param string $name
     * @return Webtatic_Form_Field
     */
    public function getField($name)
    {
        list($name, $subName) = explode('_', $name, 2);
        if ($subName) {
            return $this->_fields[$name]->getField($subName);
        } else {
            return $this->_fields[$name];
        }
    }

    public function setFields($fields)
    {
        foreach ($fields as $name => $options) {
            $this->setField($name, $options);
        }

        return $this;
    }

    /**
     * Enter description here...
     *
     * @return array
     */
    public function getFields()
    {
        return $this->_fields;
    }

    /**
     * Enter description here...
     *
     * @param string $name
     * @return Webtatic_Form
     */
    public function clearFields($name = null)
    {
        if ($name !== null) {
            unset($this->_fields[$name]);
        } else {
            $this->_fields = array();
        }
        return $this;
    }

    public function setDefaults($defaults)
    {
        foreach ($defaults as $key => $default) {
            $field = $this->getField($key);
            if ($field instanceof Webtatic_Form) {
                $field->setDefaults($default);
            } else {
                $field->setDefault($default);
            }
        }

        return $this;
    }

    /**
     * Enter description here...
     *
     * @param string|Webtatic_Form_Rule $name
     * @param array $options
     * @return Webtatic_Form
     */
    public function addRule($name, $options = null)
    {
        if (is_object($name)) {
            if ($name instanceof Webtatic_Form_Rule) {
                $this->_rules[$name->getName()] = $name;
                $name = $name->getName();
            } else {
                throw new Webtatic_Form_Exception('Object of class "'.classname($name).'" is not extended from "Webtatic_Form_Rule"');
            }
        } else {
            $this->_rules[$name] = new $this->_ruleClass($this, $name, $options);
        }
        return $this;
    }

    public function setRules($rules)
    {
        foreach ($rules as $name => $options) {
            if (is_object($options)) {
                $this->addRule($options);
            } else {
                $this->addRule($name, $options);
            }
        }
        return $this;
    }

    /**
     * Enter description here...
     *
     * @param string $name
     * @return Webtatic_Form_Rule
     */
    public function getRule($name)
    {
        return $this->_rules[$name];
    }

    /**
     * Enter description here...
     *
     * @return array
     */
    public function getRules()
    {
        return $this->_rules;
    }

    /**
     * Enter description here...
     *
     * @param string $name
     * @return Webtatic_Form
     */
    public function clearRule($name = null)
    {
        if ($name !== null) {
            unset($this->_rules[$name]);
        } else {
            $this->_rules = array();
        }
        return $this;
    }

    public function testConditions()
    {
        foreach ((array) $this->_conditions as $rule) {
            if (is_bool($rule)) {
                if (!$rule) {
                    return false;
                }
            } else {
                foreach ($rule->getFields() as $fieldName) {
                    if ($fieldName[0] == '.') {
                        $data = $this->getForm()-getField(substr($fieldName, 1));
                    } else {
                        $data = $this->getField($fieldName);
                    }
                    $args[] = $this->getRawValue();
                }

                $rule->process($args);

                if (!$rule->isValid()) {
                    return false;
                }
            }
        }

        return true;
    }

    public function process($partial = false)
    {
        $valid = true;
        $errors = array();
        $messages = array();

        if ($this->testConditions()) {
            foreach ((array) $this->getFields() as $key => $field) {
                if ($field instanceof Webtatic_Form_FieldSet && $field->getMultiple()) {
                    foreach ($field->_items as $k => $v) {
                        $errors[$key][$k] = $v->getErrors($partial);
                        $messages[$key][$k] = $v->getMessages($partial);

                        if (!$errors[$key][$k]) {
                            unset($errors[$key][$k]);
                            unset($messages[$key][$k]);
                        }

                        if (!$v->isValid($partial)) {
                            $valid = false;
                        }
                    }

                    if (!$errors[$key]) {
                        unset($errors[$key]);
                        unset($messages[$key]);
                    }
                } else {
                    $errors[$key] = $field->getErrors($partial);
                    $messages[$key] = $field->getMessages($partial);
                    if (!$errors[$key]) {
                        unset($errors[$key]);
                        unset($messages[$key]);
                    }
                    if (!$field->isValid($partial)) {
                        $valid = false;
                    }
                }
            }

            foreach ($this->getRules() as $rule) {
                if (!$partial) {
                    $args = array();
                    $datas = array();
                    $firstField = null;
                    foreach ($rule->getFields() as $fieldName) {
                        if ($fieldName[0] == '.') {
                            $field = $this->getForm()->getProcessed()->getField(substr($fieldName, 1));
                        } else {
                            $field = $this->getField($fieldName);
                        }
                        $fields[] = $field;
                        $args[] = $field->getRawValue(true);

                        if ($firstField === null) {
                            $firstField = $field;
                        }
                    }

                    $result = $rule->process($args, false);

                    if (!$rule->isValid()) {
                        $valid = false;
                        foreach ($fields as $field) {
                            $field->setValid(false);
                        }

                    } else {
                        foreach ($fields as $field) {
                            $field->setRawValue(array_shift($result['input']));
                            $field->setValue(array_shift($result['output']));
                        }
                    }
                }

                if ($rule->hasRun()) {
                    if ($this->getName() == '') {
                        $ruleName = $rule->getName();
                    } else {
                        $ruleName = $this->getName() . '_' . $rule->getName();
                    }

                    $errors[$ruleName] = $rule->getErrors();
                    $messages[$ruleName] = $rule->getMessages($firstField->getLabel());

                    if (!$errors[$key]) {
                        unset($errors[$key]);
                        unset($messages[$key]);
                    }
                }
            }
        }
        $this->setValid($valid);
        $this->setErrors($errors);
        $this->addMessages($messages);
    }

    public function getInfo()
    {
        if ($this->_valid === null) {
            $this->process(false);
        }

        $info = array();
        foreach ((array) $this->_fields as $key => $field) {
            $info[$key] = $field->getInfo();
        }

        return $info;
    }
}
?>