<?php
/**
 * This file is part of Webtatic Framework.
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.webtatic.com/license/new-bsd
 *
 * @category   Webtatic
 * @package    Webtatic_Form
 * @copyright  Copyright (c) 2007-2009 Andrew Thompson. (http://www.webtatic.com)
 * @license    http://www.webtatic.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

require_once 'Webtatic/Form/Validate/Abstract.php';
require_once 'Zend/Loader.php';

class Webtatic_Form_Validate_EmailAddress extends Webtatic_Form_Validate_Abstract
{
    protected $_validator = null;

    const INVALID            = 'emailAddressInvalid';
    const INVALID_HOSTNAME   = 'emailAddressInvalidHostname';
    const INVALID_MX_RECORD  = 'emailAddressInvalidMxRecord';
    const DOT_ATOM           = 'emailAddressDotAtom';
    const QUOTED_STRING      = 'emailAddressQuotedString';
    const INVALID_LOCAL_PART = 'emailAddressInvalidLocalPart';
    const LENGTH_EXCEEDED    = 'emailAddressLengthExceeded';

    const SERVER_SOCKET   = 'mailServerInvalid';
    const NO_HELO         = 'mailServerHelo';
    const MAIL_FROM       = "mailServerFrom";
    const RCPT_TO         = "mailServerRcpt";
    const SERVER_BUSY     = "mailServerBusy";
    const SERVER_RESPONSE = "mailServerResponse";
    const SERVER_ERROR    = "mailServerError";
    const SERVER_REJECTED = "mailServerRejected";

    protected $_messageTemplates = array(
        self::SERVER_SOCKET      => "Unable to open mailserver socket",
        self::NO_HELO            => "No response to HELO",
        self::MAIL_FROM          => "No response to MAIL FROM",
        self::RCPT_TO            => "No response to RCPT TO",
        self::SERVER_BUSY        => "451 Mailserver Busy",
        self::SERVER_RESPONSE    => "Email server reponse not OK (not 250/220)",
        self::SERVER_ERROR       => "Server did not identify as SMTP (no 220 reponse from socket connection)",
        self::SERVER_REJECTED    => "User unknown in relay recipient table",
        self::INVALID            => "'%value%' is not a valid email address in the basic format local-part@hostname",
        self::INVALID_HOSTNAME   => "'%hostname%' is not a valid hostname for email address '%value%'",
        self::INVALID_MX_RECORD  => "'%hostname%' does not appear to have a valid MX record for the email address '%value%'",
        self::DOT_ATOM           => "'%localPart%' not matched against dot-atom format",
        self::QUOTED_STRING      => "'%localPart%' not matched against quoted-string format",
        self::INVALID_LOCAL_PART => "'%localPart%' is not a valid local part for email address '%value%'",
        self::LENGTH_EXCEEDED    => "'%value%' exceeds the allowed length"
    );

    public function setValidator($value)
    {
        if (is_object($value)) {
            $this->_validator = $value;
        } else {
            $args = array();

            if (is_array($value)) {
                $args = $value;
                $value = array_shift($args);
            }

            Zend_Loader::loadClass($value);

            $class = new ReflectionClass($value);

            if ($class->hasMethod('__construct')) {
                $this->_validator = $class->newInstanceArgs($args);
            } else {
                $this->_validator = $class->newInstance();
            }
        }
    }

    public function isValid($value)
    {
        if ($this->_validator == null) {
            require_once 'Zend/Validate/EmailAddress.php';
            $this->_validator = new Zend_Validate_EmailAddress(Zend_Validate_Hostname::ALLOW_DNS, true);
        }
        if (!$this->_validator->isValid($value)) {
            foreach ($this->_validator->getErrors() as $key) {
                foreach ($this->_validator->getMessageVariables() as $variableName) {
                    $params[$variableName] = $this->_validator->$variableName;
                }

                $this->_error($key, $value, $params);
            }
            return false;
        } else {
            foreach ($this->_validator->getMessageVariables() as $variableName) {
                $params[$variableName] = $this->_validator->$variableName;
            }
            return $this->checkProvider($value, $params);
        }
    }

    public function checkProvider($value, $params)
    {
        // Lookup MX record and get the mail host name...
        if (checkdnsrr($params['hostname'], 'MX')) {
            if (getmxrr($params['hostname'], $mx)) {
                // Save the MX hostname ready for Phase 3 testing...
                $mailserver = $mx;
            } else {
                 $this->_error(self::INVALID_MX_RECORD, $value);
                 return false;
            }
        } else {
            // MX record could not be found, assume domain as mail server..
            $mailserver[] = $params['hostname'];
        }
        // Establish a connection with mailserver...
        $retry = 5;
        for ($i = 0; $i < $retry; $i++) {
            foreach ($mailserver AS $dns) {
                if ($connection = @fsockopen($dns, 25)) {
                    break 2;
                } else {
                    if ($i == $retry){
                         $this->_error(self::SERVER_SOCKET, $value);
                         return false;
                    } else {
                        sleep(1);
                    }
                }
            }
        }
        // Set mailserver connection timeout...
        stream_set_timeout($connection, 2);
        // Get DNS server response...
        $response = fgets($connection, 1024);
        // Check it is an SMTP server in order to start a conversation...
        if (ereg('^220', $response) || !$response) {
            // Let it know who they are talking to...
            $retry = 2;
            $hostname = exec('hostname');
            for ($i = 0; $i < $retry; $i++) {
                if (fputs($connection, "HELO ".$hostname."\r\n")) {
                    break;
                } else {
                    if ($i == $retry) {
                         $this->_error(self::NO_HELO);
                         return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Get response...
            for ($i = 0; $i < $retry; $i++) {
                if ($response = fgets($connection, 1024)) {
                    break;
                } else {
                    if ($i == $retry){
                        $this->_error(self::NO_HELO, true);
                        return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Request that the mailserver accepts mail from us...
            for ($i = 0; $i < $retry; $i++) {
                if (fputs($connection, "MAIL FROM: <".$params['mailFrom'].">\r\n")) {
                    $from = fgets($connection, 1024);
                    if (stristr($from, '451') && $retry < 5){
                        $retry++;
                        sleep(2);
                        continue(2);
                    }
                    break;
                } else {
                    if ($i == $retry) {
                        $this->_error(self::MAIL_FROM, $value);
                        return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Get response...
            for ($i = 0; $i < $retry; $i++) {
                if ($from){
                    break;
                } else {
                    if ($i == $retry) {
                        $this->_error(self::MAIL_FROM, $value);
                        return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Request that the mailserver accepts mail from the specified address...
            for ($i = 0; $i < $retry; $i++) {
                if (fputs($connection, "RCPT TO: <".$value.">\r\n")) {
                    break;
                } else {
                    if ($i == $retry){
                            $this->_error(self::RCPT_TO, $value);
                        return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Get response...
            for ($i = 0; $i < $retry; $i++) {
                if ($to = fgets($connection, 1024)) {
                    break;
                } else {
                    if ($i == $retry) {
                        $this->_error(self::RCPT_TO, $value);
                        return false;
                    } else {
                        sleep(1);
                    }
                }
            }
            // Say goodbye and hang up on the mailserver...
            fputs($connection, "QUIT\r\n");
            fclose($connection);
            // Check for a valid response...
            if (ereg('^250|220', $from) && ereg('^250|220', $to)) {
                return true;
            } else if (ereg('451', $from) || ereg('451', $to) || $from === false || $to === false) {
//                $this->_error(self::SERVER_BUSY, $value);
                return true;
            } else {
                $this->_error(self::SERVER_REJECTED, $value);
                return false;
            }
        } else {
            $this->_error(self::SERVER_ERROR);
            return false;
        }
    }
}
?>